/*
    ext2prepare.c -- prepares an ext2 fs for an online resize
    Copyright (C) 1999 Andreas Dilger <adilger@enel.ucalgary.ca>
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  online resize patch for the linux kernel.
*/

static const char _ext2prepare_c[] = "$Id: ext2prepare.c,v 1.7 1999/10/05 19:14:30 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include <sys/stat.h>
#include "ext2.h"

#define EXT2_CHECK_DIND 1
#define EXT2_CHECK_TIND 2

char	*prog;
int	 opt_debug = 0;
int	 opt_verbose = 0;

void print_version(FILE *outfile)
{
	fprintf(outfile, "ext2prepare v%s - 1999/10/02 for EXT2FS %s\n",
		VERSION, EXT2FS_VERSION);
}

void usage(FILE *outfile)
{
	fprintf(outfile, "usage: %s device max_new_size[bkmgt]\n", prog);
	fprintf(outfile,
"       where new_size is in ext2 filesystem blocks (1k, 2k, or 4k),\n"
"       disk Blocks (512), Kilo-, Mega-, Giga-, or Terabytes\n");
	exit(1);
}

void parse_args(int *argc, char **argv[])
{
	static struct option long_options[] = {
		{"debug", 0, 0, 'd'},
		{"verbose", 0, 0, 'v'},
		{NULL, 0, 0, 0}};
	int c;
	int dummy;

	while ((c = getopt_long(*argc, *argv, "dfuv", long_options, &dummy)) != -1)
	{
		switch (c)
		{
		case 'd': opt_debug = 1; break;
		case 'v': opt_verbose = 1; break;
		default:  usage(stderr);
		}
	}

	*argc -= optind;
	*argv += optind;
}

blk_t getfreezeroedblock(struct ext2_fs *fs, blk_t newgdblocks)
{
	int i;
	blk_t offset;

	for (i = 0, offset = fs->sb.s_first_data_block; i < fs->numgroups;
	     i++, offset += fs->sb.s_blocks_per_group)
	{
		blk_t j;

		/* Don't allocate inode blocks in range where we may need
		 * to move GDT blocks later */
		for (j = fs->itoffset + fs->inodeblocks + newgdblocks;
		     j < fs->sb.s_blocks_per_group; j++)
		{
			if (!ext2_get_block_state(fs, offset + j))
			{
				struct ext2_buffer_head *bh;

				ext2_set_block_state(fs, offset + j, 1, 1);
				bh = ext2_bcreate(fs, offset + j);
				bh->dirty = 1;
				ext2_brelse(bh, 0);

				return offset + j;
			}
		}
	}

	return 0;
}

int check_inode_ind(struct ext2_fs *fs, blk_t *block, blk_t *blocks,
		    blk_t newgdblocks)
{
	if (*block == 0)
	{
		*block = getfreezeroedblock(fs, newgdblocks);
		*blocks += 2;
		if (fs->opt_debug)
		fprintf(stderr,
			"check_inode_ind: adding indirect block 0x%04x\n",
			*block);
		return 1;
	}

	return 0;
}

int check_inode_dind(struct ext2_fs *fs, blk_t *block, blk_t *blocks,
		     blk_t newgdblocks, int resblocks, int type)
{
	struct ext2_buffer_head	*bh;
	__u32			*udata;
	int			 u32perblock = fs->blocksize >> 2;
	int			 change = 0;
	int			 i;

	change |= check_inode_ind(fs, block, blocks, newgdblocks);
	bh = ext2_bread(fs, *block);
	udata = (__u32 *)bh->data;
	if (type == EXT2_CHECK_DIND)
		for (i = 0;
		     i < u32perblock && resblocks > 0;
		     i++, udata++, resblocks -= u32perblock)
			change |= check_inode_ind(fs, udata, blocks,
						  newgdblocks);
	else if (type == EXT2_CHECK_TIND)
		for (i = 0;
		     i < u32perblock && resblocks > 0;
		     i++, udata++, resblocks -= u32perblock * u32perblock)
		{
			change |= check_inode_dind(fs, udata, blocks,
					 	   newgdblocks, resblocks,
					 	   EXT2_CHECK_DIND);

		}
	bh->dirty = change;
	ext2_brelse(bh, 0);

	return change;
}

void make_inode_ind(struct ext2_fs *fs, struct ext2_inode *inode, int resblocks,
		    blk_t newgdblocks)
{
	int	u32perblock = fs->blocksize >> 2;

	if (resblocks > EXT2_NDIR_BLOCKS) {
		check_inode_ind(fs, inode->i_block + EXT2_IND_BLOCK,
				&inode->i_blocks, newgdblocks);
		resblocks -= EXT2_NDIR_BLOCKS;
	}

	if (resblocks > u32perblock)
	{
		check_inode_dind(fs, inode->i_block + EXT2_DIND_BLOCK,
				&inode->i_blocks, newgdblocks, resblocks,
				EXT2_CHECK_DIND);

		resblocks -= u32perblock;
	}

	if (resblocks > u32perblock * u32perblock)
		check_inode_dind(fs, inode->i_block + EXT2_TIND_BLOCK,
				&inode->i_blocks, newgdblocks, resblocks,
				EXT2_CHECK_TIND);
}

int create_resize_inode(struct ext2_fs *fs, blk_t newsize)
{
	int			diff;
	int			resblocks;
	int			i;
	struct ext2_inode	inode;
	int			newgroups;
	blk_t			newgdblocks;
	blk_t			block;

	newgroups = howmany(newsize - fs->sb.s_first_data_block,
			    fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc),
			      fs->blocksize);

	ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);
	if (inode.i_mode == 0 || inode.i_links_count == 0)
	{
		if (fs->opt_verbose)
			printf("creating resize inode (%03d)\n",
				EXT2_RESIZE_INO);
		inode.i_mode = S_IFREG | S_IRUSR | S_IWUSR;
		inode.i_size = 0;
		inode.i_uid = inode.i_gid = 0;
		inode.i_atime = inode.i_ctime = time(NULL);
		inode.i_dtime = 0;
		inode.i_links_count = 1;
		inode.i_blocks = 0;
		inode.i_flags = EXT2_NODUMP_FL;
		inode.osd1.linux1.l_i_reserved1 = 0;
		for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
			inode.i_block[i] = 0;
		inode.i_block[EXT2_IND_BLOCK]  = 0;
		inode.i_block[EXT2_DIND_BLOCK] = 0;
		inode.i_block[EXT2_TIND_BLOCK] = 0;
		inode.i_version = 0;
		inode.i_file_acl = 0;
		inode.i_dir_acl = 0;
		inode.i_faddr = 0;
		inode.osd2.linux2.l_i_frag = 0;
		inode.osd2.linux2.l_i_fsize = 0;
		inode.osd2.linux2.i_pad1 = 0;
		inode.osd2.linux2.l_i_reserved2[0] = 0;
		inode.osd2.linux2.l_i_reserved2[1] = 0;
	}

	/* The maximum number of blocks to reserve happens when we have
	 * half of the filesystem full.  If we are past half full, then
	 * we need less reserved blocks.
	 */
	resblocks = (newgroups - max(fs->numgroups, newgroups / 2)) *
		    (newgdblocks - max(fs->gdblocks, newgdblocks / 2));
	make_inode_ind(fs, &inode, resblocks, newgdblocks);

	diff = newgdblocks - fs->gdblocks;
	for (i = 0, block = fs->sb.s_first_data_block + fs->gdblocks + 1;
	     i < fs->numgroups; i++, block += fs->sb.s_blocks_per_group)
	{
		int j;
		for (j = 0; j < diff; j++)
		{
			if (ext2_do_inode(fs, &inode, block + j,
					  EXT2_ACTION_FIND) < 0 &&
			    ext2_do_inode(fs, &inode, block + j,
					  EXT2_ACTION_ADD) < 0)
				return 0;
		}
	}

	inode.i_mtime = time(NULL);
	ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);

	return 1;
}


int ext2_prepare_fs(struct ext2_fs *fs, blk_t maxnewsize)
{
	int status;

	if (fs->opt_verbose)
		printf("%s: prepare for %u blocks\n", prog, maxnewsize);

	if (maxnewsize <= fs->sb.s_blocks_count)
		return 1;

	if ((fs->relocator_pool =
	     (unsigned char *)malloc(ext2_relocator_pool_size << 10)) == NULL)
	{
		fprintf(stderr, "%s: error allocating relocator pool!\n", prog);
		return 0;
	}
	fs->relocator_pool_end = fs->relocator_pool +
				 (ext2_relocator_pool_size << 10);

	status = ext2_block_relocate(fs, maxnewsize);

	if (status)
		status = ext2_metadata_push(fs, maxnewsize);

	if (status)
		status = create_resize_inode(fs, maxnewsize);

	free(fs->relocator_pool);
	fs->relocator_pool = NULL;
	fs->relocator_pool_end = NULL;

	return status;
}

int main(int argc, char *argv[])
{
	struct ext2_dev_handle	*handle;
	struct ext2_fs		*fs;
	blk_t			 newsize;
	int			 newgroups;
	char			 mod = '\0';
	char			*dev;

	prog = argv[0];
	print_version(stdout);
	parse_args(&argc, &argv);

	if (argc != 2)
		usage(stderr);

	dev = argv[0];
	handle = ext2_make_dev_handle_from_file(dev);
	if (handle == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	sscanf(argv[1], "%i%c", &newsize, &mod);

	switch (mod)	/* Order of these options is important!! */
	{
		case 't':
		case 'T': newsize *= 1024; /* no break */
		case 'g':
		case 'G': newsize *= 1024; /* no break */
		case 'm':
		case 'M': newsize *= 1024; /* no break */
		case 'k':
		case 'K': /* Need to open fs to find block size */
		case  0 : break;
		case 'b':
		case 'B': newsize /= 2; break;
		default: usage(stderr);
	}

	newgroups = howmany(newsize, 8192); /* worst case 1k blocks */

	if (ext2_max_groups < newgroups)
		ext2_max_groups = newgroups;

	if ((fs = ext2_open(handle, 0)) == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	fs->opt_debug = opt_debug;
	fs->opt_verbose = opt_verbose;

	if (mod != '\0')
		newsize /= fs->blocksize / 1024;

	if (newsize >= 1<<28 && fs->blocksize == 1024)
	{
		fprintf(stderr, "%s: can't resize 1k block filesystem over "
			"256GB.\n\ttoo much space would be wasted.\n", prog);
		return 1;
	}

	if (!ext2_prepare_fs(fs, newsize))
		return 1;

	ext2_close(fs);

	return 0;
}
